@@ -31,6 +31,7 @@ from collections import namedtuple
 
 # to avoid imap servers to kill the connection after 30mn idling
 # cf https://www.imapwiki.org/ClientImplementation/Synchronization
+
 TWENTY_NINE_MINUTES = 1740
 STOP_WAIT_SERVER_PUSH = 'stop_wait_server_push'
 
@@ -41,55 +42,97 @@ IMAP4_SSL_PORT = 993
 STARTED, CONNECTED, NONAUTH, AUTH, SELECTED, LOGOUT = 'STARTED', 'CONNECTED', 'NONAUTH', 'AUTH', 'SELECTED', 'LOGOUT'
 CRLF = b'\r\n'
 
+ID_MAX_PAIRS_COUNT = 30
+ID_MAX_FIELD_LEN = 30
+ID_MAX_VALUE_LEN = 1024
+
 AllowedVersions = ('IMAP4REV1', 'IMAP4')
-Exec = Enum('Exec', 'sync async')
+Exec = Enum('Exec', 'is_sync is_async')
 Cmd = namedtuple('Cmd', 'name           valid_states                exec')
 Commands = {
-    'APPEND':       Cmd('APPEND',       (AUTH, SELECTED),           Exec.sync),
-    'AUTHENTICATE': Cmd('AUTHENTICATE', (NONAUTH,),                 Exec.sync),
-    'CAPABILITY':   Cmd('CAPABILITY',   (NONAUTH, AUTH, SELECTED),  Exec.async),
-    'CHECK':        Cmd('CHECK',        (SELECTED,),                Exec.async),
-    'CLOSE':        Cmd('CLOSE',        (SELECTED,),                Exec.sync),
-    'COMPRESS':     Cmd('COMPRESS',     (AUTH,),                    Exec.sync),
-    'COPY':         Cmd('COPY',         (SELECTED,),                Exec.async),
-    'CREATE':       Cmd('CREATE',       (AUTH, SELECTED),           Exec.async),
-    'DELETE':       Cmd('DELETE',       (AUTH, SELECTED),           Exec.async),
-    'DELETEACL':    Cmd('DELETEACL',    (AUTH, SELECTED),           Exec.async),
-    'EXAMINE':      Cmd('EXAMINE',      (AUTH, SELECTED),           Exec.sync),
-    'EXPUNGE':      Cmd('EXPUNGE',      (SELECTED,),                Exec.async),
-    'FETCH':        Cmd('FETCH',        (SELECTED,),                Exec.async),
-    'GETACL':       Cmd('GETACL',       (AUTH, SELECTED),           Exec.async),
-    'GETQUOTA':     Cmd('GETQUOTA',     (AUTH, SELECTED),           Exec.async),
-    'GETQUOTAROOT': Cmd('GETQUOTAROOT', (AUTH, SELECTED),           Exec.async),
-    'ID':           Cmd('ID',           (NONAUTH, AUTH, LOGOUT, SELECTED), Exec.async),
-    'IDLE':         Cmd('IDLE',         (SELECTED,),                Exec.sync),
-    'LIST':         Cmd('LIST',         (AUTH, SELECTED),           Exec.async),
-    'LOGIN':        Cmd('LOGIN',        (NONAUTH,),                 Exec.sync),
-    'LOGOUT':       Cmd('LOGOUT',       (NONAUTH, AUTH, LOGOUT, SELECTED), Exec.sync),
-    'LSUB':         Cmd('LSUB',         (AUTH, SELECTED),           Exec.async),
-    'MYRIGHTS':     Cmd('MYRIGHTS',     (AUTH, SELECTED),           Exec.async),
-    'NAMESPACE':    Cmd('NAMESPACE',    (AUTH, SELECTED),           Exec.async),
-    'NOOP':         Cmd('NOOP',         (NONAUTH, AUTH, SELECTED),  Exec.async),
-    'RENAME':       Cmd('RENAME',       (AUTH, SELECTED),           Exec.async),
-    'SEARCH':       Cmd('SEARCH',       (SELECTED,),                Exec.async),
-    'SELECT':       Cmd('SELECT',       (AUTH, SELECTED),           Exec.sync),
-    'SETACL':       Cmd('SETACL',       (AUTH, SELECTED),           Exec.sync),
-    'SETQUOTA':     Cmd('SETQUOTA',     (AUTH, SELECTED),           Exec.sync),
-    'SORT':         Cmd('SORT',         (SELECTED,),                Exec.async),
-    'STARTTLS':     Cmd('STARTTLS',     (NONAUTH,),                 Exec.sync),
-    'STATUS':       Cmd('STATUS',       (AUTH, SELECTED),           Exec.async),
-    'STORE':        Cmd('STORE',        (SELECTED,),                Exec.async),
-    'SUBSCRIBE':    Cmd('SUBSCRIBE',    (AUTH, SELECTED),           Exec.sync),
-    'THREAD':       Cmd('THREAD',       (SELECTED,),                Exec.async),
-    'UID':          Cmd('UID',          (SELECTED,),                Exec.async),
-    'UNSUBSCRIBE':  Cmd('UNSUBSCRIBE',  (AUTH, SELECTED),           Exec.sync),
+    'APPEND':       Cmd('APPEND',       (AUTH, SELECTED),           Exec.is_sync),
+    'AUTHENTICATE': Cmd('AUTHENTICATE', (NONAUTH,),                 Exec.is_sync),
+    'CAPABILITY':   Cmd('CAPABILITY',   (NONAUTH, AUTH, SELECTED),  Exec.is_async),
+    'CHECK':        Cmd('CHECK',        (SELECTED,),                Exec.is_async),
+    'CLOSE':        Cmd('CLOSE',        (SELECTED,),                Exec.is_sync),
+    'COMPRESS':     Cmd('COMPRESS',     (AUTH,),                    Exec.is_sync),
+    'COPY':         Cmd('COPY',         (SELECTED,),                Exec.is_async),
+    'CREATE':       Cmd('CREATE',       (AUTH, SELECTED),           Exec.is_async),
+    'DELETE':       Cmd('DELETE',       (AUTH, SELECTED),           Exec.is_async),
+    'DELETEACL':    Cmd('DELETEACL',    (AUTH, SELECTED),           Exec.is_async),
+    'ENABLE':       Cmd('ENABLE',       (AUTH,),                    Exec.is_sync),
+    'EXAMINE':      Cmd('EXAMINE',      (AUTH, SELECTED),           Exec.is_sync),
+    'EXPUNGE':      Cmd('EXPUNGE',      (SELECTED,),                Exec.is_async),
+    'FETCH':        Cmd('FETCH',        (SELECTED,),                Exec.is_async),
+    'GETACL':       Cmd('GETACL',       (AUTH, SELECTED),           Exec.is_async),
+    'GETQUOTA':     Cmd('GETQUOTA',     (AUTH, SELECTED),           Exec.is_async),
+    'GETQUOTAROOT': Cmd('GETQUOTAROOT', (AUTH, SELECTED),           Exec.is_async),
+    'ID':           Cmd('ID',           (NONAUTH, AUTH, LOGOUT, SELECTED), Exec.is_async),
+    'IDLE':         Cmd('IDLE',         (SELECTED,),                Exec.is_sync),
+    'LIST':         Cmd('LIST',         (AUTH, SELECTED),           Exec.is_async),
+    'LOGIN':        Cmd('LOGIN',        (NONAUTH,),                 Exec.is_sync),
+    'LOGOUT':       Cmd('LOGOUT',       (NONAUTH, AUTH, LOGOUT, SELECTED), Exec.is_sync),
+    'LSUB':         Cmd('LSUB',         (AUTH, SELECTED),           Exec.is_async),
+    'MYRIGHTS':     Cmd('MYRIGHTS',     (AUTH, SELECTED),           Exec.is_async),
+    'MOVE':         Cmd('MOVE',         (SELECTED,),                Exec.is_sync),
+    'NAMESPACE':    Cmd('NAMESPACE',    (AUTH, SELECTED),           Exec.is_async),
+    'NOOP':         Cmd('NOOP',         (NONAUTH, AUTH, SELECTED),  Exec.is_async),
+    'RENAME':       Cmd('RENAME',       (AUTH, SELECTED),           Exec.is_async),
+    'SEARCH':       Cmd('SEARCH',       (SELECTED,),                Exec.is_async),
+    'SELECT':       Cmd('SELECT',       (AUTH, SELECTED),           Exec.is_sync),
+    'SETACL':       Cmd('SETACL',       (AUTH, SELECTED),           Exec.is_sync),
+    'SETQUOTA':     Cmd('SETQUOTA',     (AUTH, SELECTED),           Exec.is_sync),
+    'SORT':         Cmd('SORT',         (SELECTED,),                Exec.is_async),
+    'STARTTLS':     Cmd('STARTTLS',     (NONAUTH,),                 Exec.is_sync),
+    'STATUS':       Cmd('STATUS',       (AUTH, SELECTED),           Exec.is_async),
+    'STORE':        Cmd('STORE',        (SELECTED,),                Exec.is_async),
+    'SUBSCRIBE':    Cmd('SUBSCRIBE',    (AUTH, SELECTED),           Exec.is_sync),
+    'THREAD':       Cmd('THREAD',       (SELECTED,),                Exec.is_async),
+    'UID':          Cmd('UID',          (SELECTED,),                Exec.is_async),
+    'UNSUBSCRIBE':  Cmd('UNSUBSCRIBE',  (AUTH, SELECTED),           Exec.is_sync),
     # for testing
-    'DELAY':        Cmd('DELAY',        (AUTH, SELECTED),           Exec.sync),
+    'DELAY':        Cmd('DELAY',        (AUTH, SELECTED),           Exec.is_sync),
 }
 
 Response = namedtuple('Response', 'result lines')
 
 
+def quoted(arg):
+    """ Given a string, return a quoted string as per RFC 3501, section 9.
+
+        Implementation copied from https://github.com/mjs/imapclient
+        (imapclient/imapclient.py), 3-clause BSD license
+    """
+    if isinstance(arg, str):
+        arg = arg.replace('\\', '\\\\')
+        arg = arg.replace('"', '\\"')
+        q = '"'
+    else:
+        arg = arg.replace(b'\\', b'\\\\')
+        arg = arg.replace(b'"', b'\\"')
+        q = b'"'
+    return q + arg + q
+
+
+def arguments_rfs2971(**kwargs):
+    if kwargs:
+        if len(kwargs) > ID_MAX_PAIRS_COUNT:
+            raise ValueError('Must not send more than 30 field-value pairs')
+        args = ['(']
+        for field, value in kwargs.items():
+            field = quoted(str(field))
+            value = quoted(str(value)) if value is not None else 'NIL'
+            if len(field) > ID_MAX_FIELD_LEN:
+                raise ValueError('Field: {} must not be longer than 30'.format(field))
+            if len(value) > ID_MAX_VALUE_LEN:
+                raise ValueError('Field: {} value: {} must not be longer than 1024'.format(field, value))
+            args.extend((field, value))
+        args.append(')')
+    else:
+        args = ['NIL']
+    return args
+
+
 class Command(object):
     def __init__(self, name, tag, *args, prefix=None, untagged_resp_name=None, loop=asyncio.get_event_loop(), timeout=None):
         self.name = name
@@ -130,6 +173,9 @@ class Command(object):
     def wait_literal_data(self):
         return self._expected_size != 0 and len(self._literal_data) != self._expected_size
 
+    def wait_data(self):
+        return self.wait_literal_data()
+
     def append_literal_data(self, data):
         nb_bytes_to_add = self._expected_size - len(self._literal_data)
         self._literal_data += data[0:nb_bytes_to_add]
@@ -173,6 +219,29 @@ class Command(object):
         self._set_timer()
 
 
+class FetchCommand(Command):
+    FETCH_MESSAGE_DATA_RE = re.compile(r'[0-9]+ FETCH \(')
+
+    def __init__(self, tag, *args, prefix=None, untagged_resp_name=None,
+                 loop=asyncio.get_event_loop(), timeout=None):
+        super().__init__('FETCH', tag, *args, prefix=prefix, untagged_resp_name=untagged_resp_name,
+                         loop=loop, timeout=timeout)
+
+    def wait_data(self):
+        if self.response is None:
+            return False
+        last_fetch_index = 0
+        for index, line in enumerate(self.response.lines):
+            if isinstance(line, str) and self.FETCH_MESSAGE_DATA_RE.match(line):
+                last_fetch_index = index
+        return not matched_parenthesis(''.join(filter(lambda l: isinstance(l, str),
+                                                      self.response.lines[last_fetch_index:])))
+
+
+def matched_parenthesis(string):
+    return string.count('(') == string.count(')')
+
+
 class IdleCommand(Command):
     def __init__(self, tag, queue, *args, prefix=None, untagged_resp_name=None,
                  loop=asyncio.get_event_loop(), timeout=None):
@@ -188,11 +257,17 @@ class IdleCommand(Command):
             self.buffer.append(line)
 
     def flush(self):
-        self.queue.put_nowait(copy(self.buffer))
-        self.buffer.clear()
+        if self.buffer:
+            self.queue.put_nowait(copy(self.buffer))
+            self.buffer.clear()
 
 
-class Error(Exception):
+class AioImapException(Exception):
+    def __init__(self, reason):
+        super().__init__(reason)
+
+
+class Error(AioImapException):
     def __init__(self, reason):
         super().__init__(reason)
 
@@ -202,14 +277,15 @@ class Abort(Error):
         super().__init__(reason)
 
 
-class CommandTimeout(asyncio.TimeoutError):
+class CommandTimeout(AioImapException):
     def __init__(self, command):
         self.command = command
 
 
-class IncompleteLiteral(Exception):
-    def __init__(self, cmd):
+class IncompleteRead(AioImapException):
+    def __init__(self, cmd, data=b''):
         self.cmd = cmd
+        self.data = data
 
 
 def change_state(coro):
@@ -233,12 +309,12 @@ tagged_status_response_re = re.compile(r'[A-Z0-9]+ ((OK)|(NO)|(BAD))')
 
 
 class IMAP4ClientProtocol(asyncio.Protocol):
-    def __init__(self, loop):
+    def __init__(self, loop, conn_lost_cb=None):
         self.loop = loop
         self.transport = None
         self.state = STARTED
         self.state_condition = asyncio.Condition()
-        self.capabilities = ()
+        self.capabilities = set()
         self.pending_async_commands = dict()
         self.pending_sync_command = None
         self.idle_queue = asyncio.Queue()
@@ -246,6 +322,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         self.literal_data = None
         self.incomplete_line = b''
         self.current_command = None
+        self.conn_lost_cb = conn_lost_cb
 
         self.tagnum = 0
         self.tagpre = int2ap(random.randint(4096, 65535))
@@ -260,25 +337,31 @@ class IMAP4ClientProtocol(asyncio.Protocol):
             self._handle_responses(self.incomplete_line + d, self._handle_line, self.current_command)
             self.incomplete_line = b''
             self.current_command = None
-        except IncompleteLiteral as incomplete_literal:
-            self.current_command = incomplete_literal.cmd
-        except asyncio.IncompleteReadError as incomplete_read:
-            log.debug('Incomplete line, storing partial : %s' % incomplete_read.partial[:100])
-            self.incomplete_line = incomplete_read.partial
+        except IncompleteRead as incomplete_read:
+            self.current_command = incomplete_read.cmd
+            self.incomplete_line = incomplete_read.data
+
+    def connection_lost(self, exc):
+        log.debug('connection lost: %s', exc)
+        if self.conn_lost_cb is not None:
+            self.conn_lost_cb(exc)
 
     def _handle_responses(self, data, line_handler, current_cmd=None):
         if not data:
             if self.pending_sync_command is not None:
                 self.pending_sync_command.flush()
+            if current_cmd is not None and current_cmd.wait_data():
+                raise IncompleteRead(current_cmd)
             return
+
         if current_cmd is not None and current_cmd.wait_literal_data():
             data = current_cmd.append_literal_data(data)
             if current_cmd.wait_literal_data():
-                raise IncompleteLiteral(current_cmd)
+                raise IncompleteRead(current_cmd)
 
         line, separator, tail = data.partition(CRLF)
         if not separator:
-            raise asyncio.IncompleteReadError(data, b'line should end with CRLF')
+            raise IncompleteRead(current_cmd, data)
 
         cmd = line_handler(line.decode(), current_cmd)
 
@@ -289,6 +372,8 @@ class IMAP4ClientProtocol(asyncio.Protocol):
                 cmd = Command('NIL', 'unused')
             cmd.begin_literal_data(size)
             self._handle_responses(tail, line_handler, current_cmd=cmd)
+        elif cmd is not None and cmd.wait_data():
+            self._handle_responses(tail, line_handler, current_cmd=cmd)
         else:
             self._handle_responses(tail, line_handler)
 
@@ -297,22 +382,19 @@ class IMAP4ClientProtocol(asyncio.Protocol):
             return
 
         if self.state == CONNECTED:
-            asyncio.async(self.welcome(line))
+            asyncio.ensure_future(self.welcome(line))
         elif tagged_status_response_re.match(line):
             self._response_done(line)
         elif current_cmd is not None:
             current_cmd.append_to_resp(line)
             return current_cmd
         elif line.startswith('*'):
-            return self._untagged_response(line.replace('* ', ''))
+            return self._untagged_response(line)
         elif line.startswith('+'):
-            self._continuation(line.replace('+ ', ''))
+            self._continuation(line)
         else:
             log.info('unknown data received %s' % line)
 
-    def connection_lost(self, exc):
-        self.transport.close()
-
     def send(self, line):
         data = ('%s\r\n' % line).encode()
         log.debug('Sending : %s' % data)
@@ -326,7 +408,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         if self.pending_sync_command is not None:
             yield from self.pending_sync_command.wait()
 
-        if Commands.get(command.name).exec == Exec.sync:
+        if Commands.get(command.name).exec == Exec.is_sync:
             if self.pending_async_commands:
                 yield from self.wait_async_pending_commands()
             self.pending_sync_command = command
@@ -339,7 +421,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         try:
             yield from command.wait()
         except CommandTimeout:
-            if Commands.get(command.name).exec == Exec.sync:
+            if Commands.get(command.name).exec == Exec.is_sync:
                 self.pending_sync_command = None
             else:
                 self.pending_async_commands.pop(command.untagged_resp_name, None)
@@ -362,10 +444,13 @@ class IMAP4ClientProtocol(asyncio.Protocol):
     @asyncio.coroutine
     def login(self, user, password):
         response = yield from self.execute(
-            Command('LOGIN', self.new_tag(), user, '"%s"' % password, loop=self.loop))
+            Command('LOGIN', self.new_tag(), user, '%s' % quoted(password), loop=self.loop))
 
         if 'OK' == response.result:
             self.state = AUTH
+            for line in response.lines:
+                if 'CAPABILITY' in line:
+                    self.capabilities = self.capabilities.union(set(line.replace('CAPABILITY', '').strip().split()))
         return response
 
     @change_state
@@ -397,7 +482,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
     @asyncio.coroutine
     def idle(self):
         if 'IDLE' not in self.capabilities:
-            Abort('server has not IDLE capability')
+            raise Abort('server has not IDLE capability')
         return (yield from self.execute(IdleCommand(self.new_tag(), self.idle_queue, loop=self.loop)))
 
     def has_pending_idle_command(self):
@@ -417,8 +502,8 @@ class IMAP4ClientProtocol(asyncio.Protocol):
     @asyncio.coroutine
     def fetch(self, message_set, message_parts, by_uid=False, timeout=None):
         return (yield from self.execute(
-            Command('FETCH', self.new_tag(), message_set, message_parts,
-                    prefix='UID' if by_uid else '', loop=self.loop, timeout=timeout)))
+            FetchCommand(self.new_tag(), message_set, message_parts,
+                         prefix='UID' if by_uid else '', loop=self.loop, timeout=timeout)))
 
     @asyncio.coroutine
     def store(self, *args, by_uid=False):
@@ -427,41 +512,57 @@ class IMAP4ClientProtocol(asyncio.Protocol):
                     prefix='UID' if by_uid else '', untagged_resp_name='FETCH', loop=self.loop)))
 
     @asyncio.coroutine
-    def expunge(self):
-        return (yield from self.execute(Command('EXPUNGE', self.new_tag(), loop=self.loop)))
+    def expunge(self, *args, by_uid=False):
+        return (yield from self.execute(
+            Command('EXPUNGE', self.new_tag(), *args,
+                    prefix='UID' if by_uid else '', loop=self.loop)))
 
     @asyncio.coroutine
-    def uid(self, command, *criteria):
+    def uid(self, command, *criteria, timeout=None):
         if self.state not in Commands.get('UID').valid_states:
             raise Abort('command UID illegal in state %s' % self.state)
 
         if command.upper() == 'FETCH':
-            return (yield from self.fetch(criteria[0], criteria[1], by_uid=True))
+            return (yield from self.fetch(criteria[0], criteria[1], by_uid=True, timeout=timeout))
         if command.upper() == 'STORE':
             return (yield from self.store(*criteria, by_uid=True))
         if command.upper() == 'COPY':
             return (yield from self.copy(*criteria, by_uid=True))
-        else:
-            raise Abort('command UID only possible with COPY, FETCH or STORE (was %s)' % command.upper())
+        if command.upper() == 'MOVE':
+            return (yield from self.move(*criteria, by_uid=True))
+        if command.upper() == 'EXPUNGE':
+            if 'UIDPLUS' not in self.capabilities:
+                raise Abort('EXPUNGE with uids is only valid with UIDPLUS capability. UIDPLUS not in (%s)' % self.capabilities)
+            return (yield from self.expunge(*criteria, by_uid=True))
+        raise Abort('command UID only possible with COPY, FETCH, EXPUNGE (w/UIDPLUS) or STORE (was %s)' % command.upper())
 
     @asyncio.coroutine
-    def copy(self, *args, by_uid=True):
+    def copy(self, *args, by_uid=False):
         return (yield from self.execute(
             Command('COPY', self.new_tag(), *args, prefix='UID' if by_uid else '', loop=self.loop)))
 
+    @asyncio.coroutine
+    def move(self, uid_set, mailbox, by_uid=False):
+        if 'MOVE' not in self.capabilities:
+            raise Abort('server has not MOVE capability')
+
+        return (yield from self.execute(
+            Command('MOVE', self.new_tag(), uid_set, mailbox, prefix='UID' if by_uid else '', loop=self.loop)))
+
     @asyncio.coroutine
     def capability(self):
         response = yield from self.execute(Command('CAPABILITY', self.new_tag(), loop=self.loop))
 
-        self.capabilities = response.lines[0].split()
-        version = self.capabilities[0].upper()
+        capability_list = response.lines[0].split()
+        self.capabilities = set(capability_list)
+        version = capability_list[0].upper()
         if version not in AllowedVersions:
             raise Error('server not IMAP4 compliant')
         else:
             self.imap_version = version
 
     @asyncio.coroutine
-    def append(self, message_bytes, mailbox='INBOX', flags=None, date=None):
+    def append(self, message_bytes, mailbox='INBOX', flags=None, date=None, timeout=None):
         args = [mailbox]
         if flags is not None:
             if (flags[0], flags[-1]) != ('(', ')'):
@@ -472,10 +573,21 @@ class IMAP4ClientProtocol(asyncio.Protocol):
             args.append(time2internaldate(date))
         args.append('{%s}' % len(message_bytes))
         self.literal_data = message_bytes
-        return (yield from self.execute(Command('APPEND', self.new_tag(), *args, loop=self.loop)))
+        return (yield from self.execute(Command('APPEND', self.new_tag(), *args, loop=self.loop, timeout=timeout)))
+
+    @asyncio.coroutine
+    def id(self, **kwargs):
+        args = arguments_rfs2971(**kwargs)
+        return (yield from self.execute(Command('ID', self.new_tag(), *args, loop=self.loop)))
 
     simple_commands = {'NOOP', 'CHECK', 'STATUS', 'CREATE', 'DELETE', 'RENAME',
-                       'SUBSCRIBE', 'UNSUBSCRIBE', 'LSUB', 'LIST', 'EXAMINE'}
+                       'SUBSCRIBE', 'UNSUBSCRIBE', 'LSUB', 'LIST', 'EXAMINE', 'ENABLE'}
+
+    @asyncio.coroutine
+    def namespace(self):
+        if 'NAMESPACE' not in self.capabilities:
+            raise Abort('server has not NAMESPACE capability')
+        return (yield from self.execute(Command('NAMESPACE', self.new_tag(), loop=self.loop)))
 
     @asyncio.coroutine
     def simple_command(self, name, *args):
@@ -485,7 +597,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
 
     @asyncio.coroutine
     def wait_async_pending_commands(self):
-        yield from asyncio.wait([asyncio.async(cmd.wait()) for cmd in self.pending_async_commands.values()])
+        yield from asyncio.wait([asyncio.ensure_future(cmd.wait()) for cmd in self.pending_async_commands.values()])
 
     @asyncio.coroutine
     def wait(self, state_regexp):
@@ -494,6 +606,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
             yield from self.state_condition.wait_for(lambda: state_re.match(self.state))
 
     def _untagged_response(self, line):
+        line = line.replace('* ', '')
         if self.pending_sync_command is not None:
             self.pending_sync_command.append_to_resp(line)
             command = self.pending_sync_command
@@ -538,8 +651,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         command.close(response_text, result=response_result)
 
     def _continuation(self, line):
-        if 'literal data' in line:
-            # APPEND case
+        if self.pending_sync_command is not None and self.pending_sync_command.name == 'APPEND':
             if self.literal_data is None:
                 Abort('asked for literal data but have no literal data to send')
             self.transport.write(self.literal_data)
@@ -548,6 +660,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         elif self.pending_sync_command is not None:
             log.debug('continuation line appended to pending sync command %s : %s' % (self.pending_sync_command, line))
             self.pending_sync_command.append_to_resp(line)
+            self.pending_sync_command.flush()
         else:
             log.info('server says %s (ignored)' % line)
 
@@ -563,16 +676,20 @@ class IMAP4ClientProtocol(asyncio.Protocol):
 class IMAP4(object):
     TIMEOUT_SECONDS = 10
 
-    def __init__(self, host='localhost', port=IMAP4_PORT, loop=asyncio.get_event_loop(), timeout=TIMEOUT_SECONDS):
+    def __init__(self, host='127.0.0.1', port=IMAP4_PORT, loop=asyncio.get_event_loop(), timeout=TIMEOUT_SECONDS, conn_lost_cb=None, ssl_context=None):
         self.timeout = timeout
         self.port = port
         self.host = host
         self.protocol = None
-        self.create_client(host, port, loop)
+        self._idle_waiter = None
+        self.create_client(host, port, loop, conn_lost_cb, ssl_context)
 
-    def create_client(self, host, port, loop):
-        self.protocol = IMAP4ClientProtocol(loop)
-        loop.create_task(loop.create_connection(lambda: self.protocol, host, port))
+    def create_client(self, host, port, loop, conn_lost_cb=None, ssl_context=None):
+        self.protocol = IMAP4ClientProtocol(loop, conn_lost_cb)
+        loop.create_task(loop.create_connection(lambda: self.protocol, host, port, ssl=ssl_context))
+
+    def get_state(self):
+        return self.protocol.state
 
     @asyncio.coroutine
     def wait_hello_from_server(self):
@@ -601,7 +718,7 @@ class IMAP4(object):
 
     @asyncio.coroutine
     def uid(self, command, *criteria):
-        return (yield from asyncio.wait_for(self.protocol.uid(command, *criteria), self.timeout))
+        return (yield from self.protocol.uid(command, *criteria, timeout=self.timeout))
 
     @asyncio.coroutine
     def store(self, *criteria):
@@ -624,6 +741,8 @@ class IMAP4(object):
         return (yield from self.protocol.idle())
 
     def idle_done(self):
+        if self._idle_waiter is not None:
+            self._idle_waiter.cancel()
         self.protocol.idle_done()
 
     @asyncio.coroutine
@@ -639,18 +758,24 @@ class IMAP4(object):
 
     @asyncio.coroutine
     def idle_start(self, timeout=TWENTY_NINE_MINUTES):
-        idle = asyncio.async(self.idle())
-        self.protocol.loop.call_later(timeout, lambda: asyncio.async(self.stop_wait_server_push()))
-        yield from self.wait_idled()
+        if self._idle_waiter is not None:
+            self._idle_waiter.cancel()
+        idle = asyncio.ensure_future(self.idle())
+        self._idle_waiter = self.protocol.loop.call_later(timeout, lambda: asyncio.ensure_future(self.stop_wait_server_push()))
+        yield from self.wait_server_push(self.timeout) # idling continuation
         return idle
 
-    @asyncio.coroutine
-    def wait_idled(self):
-        yield from self.wait_server_push(self.timeout)
-
     def has_pending_idle(self):
         return self.protocol.has_pending_idle_command()
 
+    @asyncio.coroutine
+    def id(self, **kwargs):
+        return (yield from asyncio.wait_for(self.protocol.id(**kwargs), self.timeout))
+
+    @asyncio.coroutine
+    def namespace(self):
+        return (yield from asyncio.wait_for(self.protocol.namespace(), self.timeout))
+
     @asyncio.coroutine
     def noop(self):
         return (yield from asyncio.wait_for(self.protocol.simple_command('NOOP'), self.timeout))
@@ -697,12 +822,26 @@ class IMAP4(object):
 
     @asyncio.coroutine
     def append(self, message_bytes, mailbox='INBOX', flags=None, date=None):
-        return (yield from asyncio.wait_for(self.protocol.append(message_bytes, mailbox, flags, date), self.timeout))
+        return (yield from self.protocol.append(message_bytes, mailbox, flags, date, timeout=self.timeout))
 
     @asyncio.coroutine
     def close(self):
         return (yield from asyncio.wait_for(self.protocol.close(), self.timeout))
 
+    @asyncio.coroutine
+    def move(self, uid_set, mailbox):
+        return (yield from asyncio.wait_for(self.protocol.move(uid_set, mailbox), self.timeout))
+
+    @asyncio.coroutine
+    def enable(self, capability):
+        if 'ENABLE' not in self.protocol.capabilities:
+            raise Abort('server has not ENABLE capability')
+
+        return (yield from asyncio.wait_for(self.protocol.simple_command('ENABLE', capability), self.timeout))
+
+    def has_capability(self, capability):
+        return capability in self.protocol.capabilities
+
 
 def extract_exists(response):
     for line in response.lines:
@@ -711,14 +850,14 @@ def extract_exists(response):
 
 
 class IMAP4_SSL(IMAP4):
-    def __init__(self, host='localhost', port=IMAP4_SSL_PORT, loop=asyncio.get_event_loop(),
-                 timeout=IMAP4.TIMEOUT_SECONDS):
-        super().__init__(host, port, loop, timeout)
+    def __init__(self, host='127.0.0.1', port=IMAP4_SSL_PORT, loop=asyncio.get_event_loop(),
+                 timeout=IMAP4.TIMEOUT_SECONDS, ssl_context=None):
+        super().__init__(host, port, loop, timeout, None, ssl_context)
 
-    def create_client(self, host, port, loop):
-        self.protocol = IMAP4ClientProtocol(loop)
-        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
-        loop.create_task(loop.create_connection(lambda: self.protocol, host, port, ssl=ssl_context))
+    def create_client(self, host, port, loop, conn_lost_cb=None, ssl_context=None):
+        if ssl_context is None:
+            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
+        super().create_client(host, port, loop, conn_lost_cb, ssl_context)
 
 
 # functions from imaplib
